from dunamai import Version
import subprocess
import toml


def _check_clean_git():
    status_output = subprocess.run(
        ["git", "status", "--porcelain"],
        capture_output=True,
        text=True,
        check=True,
    )
    changes = status_output.stdout.splitlines()
    if len(changes) > 0:
        print(f"Following files have changes {changes}")
        print("Ensure git tree is clean")
        SystemExit("Exiting...")


def _update_version() -> str:
    v = Version.from_git()
    print(f"Current version: {v.base}")
    print("How would you like to bum the version?")
    print("  [0] Major")
    print("  [1] Minor")
    print("  [2] Patch (default)")
    print("  [3] Manual")
    bump_type = input("")

    if bump_type == "":
        bump_type = "2"

    if bump_type not in {"0", "1", "2", "3"}:
        raise ValueError(f"Incorrect selection {bump_type}")
    elif bump_type == "3":
        v_new = input("Enter new version in format {Major}.{Minor}.{Patch}:\n")
        return v_new

    bump_type = int(bump_type)
    v_list = [int(x) for x in v.base.split(".")]
    v_list[bump_type] += 1

    if bump_type < 2:
        for x in range(bump_type, 2):
            v_list[x + 1] = 0

    v_new = ".".join(str(x) for x in v_list)
    print(f"New version: {v_new}")
    return v_new


def _update_pyproject_toml(v_new: str):
    with open("pyproject.toml", "r") as f:
        data = toml.load(f)

    data["project"]["version"] = v_new

    with open("pyproject.toml", "w") as f:
        toml.dump(data, f)


def _git_commit_and_tag(v_new: str):
    """Commit version update, create a Git tag, and push it."""
    result = subprocess.run(
        ["git", "log", "-1", "--pretty=%B"], capture_output=True, text=True, check=True
    )
    current_message = result.stdout.strip()
    try:
        subprocess.run(["git", "add", "pyproject.toml"], check=True)
        subprocess.run(
            [
                "git",
                "commit",
                "--amend",
                "-m",
                f"{current_message}\n\nversion({v_new})",
            ],
            check=True,
        )
        subprocess.run(["git", "tag", f"v{v_new}"], check=True)
        subprocess.run(["git", "push", "--force-with-lease"], check=True)
        subprocess.run(["git", "push", "--tags"], check=True)
        print(f"Successfully tagged and pushed version {v_new}.")
    except subprocess.CalledProcessError as e:
        print(f"Error running git command: {e}")
        SystemExit("Exiting...")


def bump():
    _check_clean_git()
    v_new = _update_version()
    _update_pyproject_toml(v_new)
    _git_commit_and_tag(v_new)



from typing import Optional, Sequence, Type
import dagster as dg
from dagster_duckdb.io_manager import DuckDBIOManager, DuckDbClient

from dagster_db.io_managers.custom_db_io_manager import CustomDbIOManager
from dagster_db.type_handlers.custom_type_handler import CustomDbTypeHandler


def build_custom_duckdb_io_manager(
    type_handlers: Sequence[CustomDbTypeHandler],
    default_load_type: Optional[Type] = None,
    io_manager_name: str = "DuckDbIoManager"
) -> dg.IOManagerDefinition:
    @dg.io_manager(config_schema=DuckDBIOManager.to_config_schema())
    def duckdb_io_manager(init_context):
        return CustomDbIOManager(
            type_handlers=type_handlers,
            db_client=DuckDbClient(),
            io_manager_name=io_manager_name,
            database=init_context.resource_config["database"],
            schema=init_context.resource_config.get("schema"),
            default_load_type=default_load_type,
        )

    return duckdb_io_manager



import dagster as dg
from typing import Sequence, Type
from dagster._core.storage.db_io_manager import TableSlice
import pandas as pd
from duckdb import BinderException, DuckDBPyConnection, IOException
from dagster_duckdb.io_manager import DuckDbClient
from dagster._utils.backoff import backoff

from dagster_db.helpers.db import table_slice_to_schema_table
from dagster_db.helpers.pandas import get_sample_md, get_table_schema, glimpse
from dagster_db.type_handlers.custom_type_handler import CustomDbTypeHandler

class DuckDbPandasTypeHandler(CustomDbTypeHandler[pd.DataFrame, DuckDBPyConnection]):

    @property
    def supported_types(self) -> Sequence[Type[object]]:
        return [pd.DataFrame]

    def validate_obj_db(self, context, obj_db: pd.DataFrame, connection: DuckDBPyConnection):
        return

    def db_safe_transformations(self, context, obj: pd.DataFrame, connection: DuckDBPyConnection) -> pd.DataFrame:
        return obj

    def output_metadata(self, context: dg.OutputContext, obj: pd.DataFrame, obj_db: pd.DataFrame):
        return {
            "sample_obj": dg.MarkdownMetadataValue(get_sample_md(obj)),
            "sample_obj_db": dg.MarkdownMetadataValue(get_sample_md(obj_db)),
            "rows": dg.IntMetadataValue(obj.shape[0]),
            "table_schema": dg.TableSchemaMetadataValue(get_table_schema(obj_db)),
            }

    @classmethod
    def _load_into_db(table_schema, obj: pd.DataFrame, connection: DuckDBPyConnection):
        connection.execute(f"CREATE TABLE IF NOT EXISTS {table_schema} as SELECT * FROM obj;")
        if not connection.fetchall():
            connection.execute(f"INSERT INTO {table_schema} SELECT * FROM obj;")


    def handle_output(self, context: dg.OutputContext, table_slice: TableSlice, obj: pd.DataFrame, connection: DuckDBPyConnection):
        table_schema = table_slice_to_schema_table(table_slice)
        connection.execute(f"CREATE SCHEMA IF NOT EXISTS {table_slice.schema};")
        connection.execute("SET GLOBAL pandas_analyze_sample=10000")
        try:
            backoff(
                self._load_into_db,
                retry_on=(IOException,),
                kwargs={
                    "table_schema": table_schema,
                    "obj": obj,
                    "connection": connection,
                },
                max_retries=1,
            )
        except BinderException:
            obj_existing = self.load_input(context, table_slice, connection)
            msg = f"""`obj` incompatible with existing table

            `obj`
            {glimpse(obj)}

            `obj_existing`
            {glimpse(obj_existing)}
            """
            context.log.error(msg)
        return

    def load_input(self, context: dg.InputContext | dg.OutputContext, table_slice: TableSlice, connection: DuckDBPyConnection) -> pd.DataFrame:
        if table_slice.partition_dimensions and len(context.asset_partition_keys) == 0:
            raise ValueError(f"{table_slice.partition_dimensions=} incompatible with {context.asset_partition_keys=}")

        query = DuckDbClient.get_select_statement(table_slice)
        result = connection.execute(query=query)
        obj = result.df()
        glimpse(obj)
        return obj



from typing import Optional
import pandas as pd
import numpy as np
import dagster as dg

def get_sample_md(obj: pd.DataFrame, n_max=10) -> Optional[str]:
    return obj.sample(n=min(n_max, obj.shape[0])).astype("string").fillna("").to_markdown()


def get_summary_md(obj: pd.DataFrame) -> Optional[str]:
    return obj.describe(include="all", exclude=["O", "B"]).fillna(np.inf).to_markdown()


def get_table_schema(
    obj: pd.DataFrame
) -> dg.TableSchema:
    return dg.TableSchema(
        columns=[
            dg.TableColumn(name=name, type=str(dtype))
            for name, dtype in obj.dtypes.items()
        ]
    )


def glimpse(df: pd.DataFrame) -> str:
    string = f"""Rows: {df.shape[0]}\nColumns: {df.shape[1]}\n\n"""
    for col in df.columns:
        string += f"$ {col} <{df[col].dtype}> {df[col].head().to_numpy()}\n"

    return string

